// averages.cpp
//
// mwanova - Multi-Way Analysis of Variance.
// Copyright (C) 2001  Antonio Santos (amsantos@fc.up.pt)
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// *************************************************************************
// 

#include <iostream.h>
#include <string.h>
#include <stdio.h>
#include <iomanip.h>
#include <math.h>
#include "data.h"
#include "probs.h"

void data::get_averages(CODES cline, const char *fname, int order, double err)
{ 
 partial *t,*firstp,*lastp,*p;
 bool    equal;
 int     i;
 
 firstp=NULL;
 lastp=NULL;
  
 if(first){
  cout << "Averages for " << fname << endl;
  footer();
  // Create a list of averages
  
  t=first;
  do{   
   if(!firstp){           // No partials in the list, add first
    firstp = new partial;
    if(firstp){
     for(i=0;i<MAXFACTORS;i++){
      if(cline[i]>0) firstp->codes[i]=t->orig[i];
      else firstp->codes[i]=0;
     }
     firstp->sum=t->sum;
     firstp->sum2=t->sum2;
     firstp->n=t->n;
     firstp->next=NULL;
     lastp=firstp;
    }
   }
   else{                  // There are partials in the list...
    p=firstp;    
    do{
     equal=true;
     for(i=0;i<MAXFACTORS;i++){
      if(cline[i]>0) if((p->codes[i])!=(t->orig[i])) {equal=false; break;}
     }
     if(!equal) p=p->next;
    }while((p)&&(!equal));
    
    if(equal){
     
     // There is one item equal to 't->codes'! Add sums and n values to it
     
     p->sum+=t->sum;
     p->sum2+=t->sum2;
     p->n+=t->n;
    }
    else{
     // No item in the list is equal to current 't->codes'. 
     //Add a new one.
     
     p = new partial;
     if(p){
      for(i=0;i<MAXFACTORS;i++){
       if(cline[i]>0) p->codes[i]=t->orig[i];
       else p->codes[i]=0;
      }
      p->sum=t->sum;
      p->sum2=t->sum2;
      p->n=t->n;
      p->next=NULL;
      lastp->next=p;
      lastp=p;
     }
    } 
   }
   t=t->next;
  }while(t);
  
  if(firstp){
  
   // Compute variances
   
   p=firstp;
   do{
    p->var=(p->sum2-pow(p->sum,2)/p->n)/(p->n-1);
    p=p->next;
   }while(p);
   
   // Output averages
   for(i=0;i<get_factors();i++){
    if(cline[i]>0) cout << setw(4) << get_factor_name(i);
   }
   cout << setw(6) << "n";
   cout << setw(17) << "Average" << setw(17) << "Variance" << endl;
   footer();
   p=firstp;
   do{
    for(i=0;i<get_factors();i++){
     if(cline[i]>0) cout << setw(4) << get_orig_code_name(i,p->codes[i]);
    } 
    cout << setw(6) << p->n;
    cout << setw(17) << p->sum/p->n << setw(17) << p->var << endl;
    p=p->next;
   }while(p); 
   footer();
   
   // Do multiple tests if term is less than 2nd order interactions 
   
   if(order<4) multi_tests(firstp, lastp, err);
   
   // Destruct list of averages
   
   do{
    p=firstp;
    firstp=p->next;
    delete p;
   }while(firstp); 
  }
 }
}

void data::multi_tests(partial *firstp, partial *lastp, double err)
{
 struct average{
  double  mean;
  int     n;
  int     ord;
  average *next;
  average *prev;
 };
 
 partial *p;
 average *firstavg,*lastavg,*a1,*a2;
 double  a,q,se;
 bool    inserted;
 int     ord,nmeans;
 char    temp[100];
 
 firstavg=NULL;
 lastavg=NULL;
 ord=0;
 
 if(err==0) return;
 
 if(firstp){
 
  cout << "Multiple tests" << endl;
 
  // Sort averages  
  p=firstp;
  do{
   if(!firstavg){
    a1 = new average;
    if(a1){
     a1->mean=p->sum/p->n;
     a1->n=p->n;
     a1->next=NULL;
     a1->prev=NULL;
     firstavg=a1;
     lastavg=a1;
    } 
   }
   else{
    inserted=false;
    a=p->sum/p->n;
    a1=firstavg;
    do{
     if(a<a1->mean){
      if(a1==firstavg){
       a2 = new average;
       if(a2){
        a2->next=firstavg;
	a2->prev=NULL;
	firstavg->prev=a2;
	a2->mean=a;
	a2->n=p->n;
	firstavg=a2;
	inserted=true;
       }
      }
      else{
       a2 = new average;
       if(a2){
        a2->mean=a;
	a2->n=p->n;
	a2->next=a1;
	a2->prev=a1->prev;
	a1->prev->next=a2;
	a1->prev=a2;
	inserted=true;
       }
      }
     }
     a1=a1->next;
    }while(a1&&(!inserted));
    
    if(!inserted){
     a1=new average;
     a1->mean=a;
     a1->n=p->n;
     a1->next=NULL;
     a1->prev=lastavg;
     lastavg->next=a1;
     lastavg=a1;
    }
   }
   p=p->next;
  }while(p);
  
  if(firstavg){
  
   // Give an order to all members
   
   ord=0;
   a1=firstavg;
   do{
    ord++;
    a1->ord=ord;
    a1=a1->next;
   }while(a1);
   
   // Compute SNK tests
   
   footer();
   cout << setw(8) << "Average ";
   cout << "| ";
   a1=firstavg;
   do{
    cout << setw(10) << a1->mean;
    a1=a1->next;
   }while(a1->next);
   cout << endl;
   footer();
   
   a1=firstavg->next;
   do{
    cout << setw(8) << a1->mean;
    cout << "| ";
    a2=firstavg;
    do{
     nmeans=abs(a1->ord-a2->ord+1);
     q=fabs(a1->mean-a2->mean);
     se=sqrt(err/(a1->n-1));
     q/=se;
     cout << setw(10) << q;    
     a2=a2->next;
    }while(a2!=a1);
    cout << endl;
    
    // Repeat to compute probabilities
    
    cout << setw(8) << " " << "| ";
    a2=firstavg;
    do{
     nmeans=abs(a1->ord-a2->ord+1);
     q=fabs(a1->mean-a2->mean);
     se=sqrt(err/(a1->n-1));
     q/=se;
     sprintf(temp,"(%5.3f)",qprob(q,nmeans,a1->n-1));
     cout << setw(10) << temp;   
     a2=a2->next;
    }while(a2!=a1);
    cout << endl;
    
    a1=a1->next;
   }while(a1);
   footer();
   
   // Destruct list;
   do{
    a1=firstavg;
    firstavg=a1->next;
    delete firstavg;
   }while(firstavg);
  }
 }
}
